#include <dft.h>
#include <ddn/ddn_pt0.h>

#include <gmp.h>

#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>

#define OPT_EXT 0x7f

enum opt_ext
{
	OPT_WINSIZE = OPT_EXT,
	OPT_WINFUNC,
};

const char *prog;

static void
usage(void)
{
	fprintf(stderr, "usage: %s\n\n", prog);
	fprintf(stderr, "  -h, --help                   Show this help message.\n");
	fprintf(stderr, "  -n, --nodes <count>          Number of nodes in the network.\n");
	fprintf(stderr, "  -V, --version                Show version information.\n");
	fprintf(stderr, "      --winsize                Size of DFT window.\n");
	fprintf(stderr, "      --winfunc                Window function for DFT.\n");
}

static void
version(void)
{
	fprintf(stderr, "ddndft v0\n");
}

static const struct option opt_long[] =
{
	{ "help", 0, NULL, 'h' },
	{ "nodes", 1, NULL, 'n' },
	{ "version", 0, NULL, 'V' },
	{ "winfunc", 1, NULL, OPT_WINFUNC },
	{ "winsize", 1, NULL, OPT_WINSIZE },
	{ NULL, 0, NULL, 0 },
};

static const char opt_short[] = ":hn:V";

static size_t opt_winsize = 512;
static const char *opt_winfunc = "rect";
static double (*winfunc)(size_t, size_t);
static size_t opt_nodes;

int
main(int argc, char **argv)
{
	struct ddn_pt0_fitness_data fit;
	int ret;

	prog = basename(argv[0]);

	while (1)
	{
		int opt;

		opt = getopt_long(argc, argv, opt_short, opt_long, NULL);
		if (opt == -1)
			break;
		switch (opt)
		{
		case 'h':
			usage();
			return 0;
		case 'n':
			opt_nodes = atoi(optarg);
			if (opt_nodes == 0)
			{
				fprintf(stderr, "%s: error: zero nodes is too small a network\n",
					prog);
				return 5;
			}
			break;
		case 'V':
			version();
			return 0;
		case OPT_WINSIZE:
			/* FIXME: strtoul */
			opt_winsize = atoi(optarg);
			break;
		case OPT_WINFUNC:
			opt_winfunc = optarg;
			break;
		case '?':
			fprintf(stderr, "%s: error: unknown option\n", prog);
			return 1;
		case ':':
			fprintf(stderr, "%s: error: missing argument\n", prog);
			return 2;
		default:
			fprintf(stderr, "%s: error: weird return value '%c'\n", prog, opt);
			return 3;
		}
	}

	winfunc = dft_window_func_from_str(opt_winfunc);
	if (!winfunc)
	{
		fprintf(stderr, "%s: error: unknown window function '%s'\n",
			prog, opt_winfunc);
		return 6;
	}

	ret = ddn_pt0_fitness_init(&fit, opt_winsize, winfunc, opt_nodes);
	if (ret)
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(-ret));
		return 7;
	}

	dft_init(&fit.dft);

	while (1)
	{
		size_t count;
		int c;
		double sample;

		mpz_set_ui(fit.znum, 0);
		do
		{
			c = fgetc(stdin);
		} while (isspace(c));

		if (c == EOF)
			break;
		for (count = 0; c != EOF && !isspace(c); c = fgetc(stdin))
		{
			mpz_mul_2exp(fit.znum, fit.znum, 1);
			if (c == '1' || c == '0')
			{
				if (c == '1')
					mpz_setbit(fit.znum, 0);
				++count;
			}
			else
			{
				fprintf(stderr, "%s: error: garbage character '%c'\n", prog, c);
				return 12;
			}
		}

		if (!isspace(c) && c != EOF)
		{
			fprintf(stderr, "%s: error: garbage at the end of input\n", prog);
			return 13;
		}

		if (!opt_nodes)
		{
			opt_nodes = count;
		}

		if (count != opt_nodes)
		{
			fprintf(stderr, "%s: error: got '%zu' values, expected '%zu'\n",
				prog, count, opt_nodes);
			return 14;
		}

		sample = ddn_pt0_fitness_sample(&fit, count);
		dft_sample_add(&fit.dft, sample);
	}

	dft_finish(&fit.dft);

	for (size_t i = 0; i < fit.dft.transformed_size; ++i)
	{
		printf("%zu %f\n", i, fit.dft.result[i]);
		//* pow(-1.0, i));
	}

	ddn_pt0_fitness_cleanup(&fit);

	if (ferror(stdin))
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
		return 10;
	}

	if (fflush(stdout))
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
		return 11;
	}

	if (ferror(stdout))
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
		return 12;
	}

	return 0;
}
