#include <ddn/ddn_gt0.h>
#include <ddn/ddn_pt0.h>
#include <ga/ga.h>
#include <uprand.h>

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <stdio.h>

#define OPT_EXT 0x7f

enum opt_ext
{
	OPT_ADULTS = 'a',
	OPT_CROSSOVER_PROB = 'x',
	OPT_CHILDREN = 'c',
	OPT_ELITES = 'e',
	OPT_GENERATION_PRINT = 'G',
	OPT_GENERATIONS = 'g',
	OPT_HELP = 'h',
	OPT_LOG = 'l',
	OPT_MUTATION_PROB = 'M',
	OPT_NODES = 'n',
	OPT_CROSSOVER_POINTS = 'P',
	OPT_VERSION = 'V',
	OPT_OPTION = OPT_EXT,
	OPT_ADULT_SEL_MECH,
	OPT_ADULT_SEL_PROTO,
	OPT_PARENT_SEL_MECH,
	OPT_PARENT_SEL_PROTO,
	OPT_POPULATION_LOG,
	OPT_STATES,
};

const char *prog;

static void
usage(void)
{
	fprintf(stderr, "usage: %s\n\n", prog);
	fprintf(stderr, 
"  -a, --adults <count>            Use <count> adults in each generation.\n"
"      --adult-sel-mech  <mech>    Use <mech> as adult selection mechanism.\n"
"      --adult-sel-proto <proto>   Use <proto> as adult selection protocol.\n"
"  -C, --crossover-prob  <prob>    Have crossover probability <prob>.\n"
"  -c, --children <count>          Use <count> children in each generation.\n"
"  -e, --elites <count>            Use <count> elites in each generation.\n"
"  -G, --generation-print <count>  Output every <count> generation to file.\n"
"  -g, --generations <count>       Run for <count> generations.\n"
"  -h, --help                      Show this help message.\n"
"  -l, --log <file>                Log progress to <file>.\n"
"  -M, --mutation-prob <prob>      Have mutation probability <prob>.\n"
"  -n, --nodes <count>             The networks will have <count> nodes.\n"
"      --option <option>           Set experiment specific option.\n"
"      --parent-sel-mech  <mech>   Use <mech> as parent selection mechanism.\n"
"  -P, --crossover-points <count>  Have <count> crossover points.\n"
"      --population-log <file>     Log entire population to <file>.\n"
"      --states <file>             Dump states to <file>.\n"
"  -V, --version                   Show version information.\n"
"\n"
"  Valid options for the --adult-sel-mech option:\n"
"    best-fit                      Select individuals with highes fitness.\n"
"    stochastic                    Select individuals randomly.\n"
"\n"
"  Valid options for the --adult-sel-proto option:\n"
"    full-gen-rep                  Replace entire adult generation.\n"
"    gen-mix                       Have a mix of children and adults to select from.\n"
"\n"
"  Valid options for the --parent-sel-mech option:\n"
"    fit-prop                      Use fitness proportionate selection.\n"
"    sigma-scale                   Use sigma scaling proportionate selection.\n"
"    tournament                    Use tournament selection.\n"
"\n"
"  Valid options for the --option option:\n"
"    dft-winsize=size              Window size used with DTF.\n"
"    dft-winfunc=func              Window function used with DFT.\n"
"    sim-statecount=count          The number of states to step through when simulating the DDN.\n"
		);
}

static void
version(void)
{
	fprintf(stderr, "ddnevo v0\n");
}

static size_t opt_adults = 8;
static char *opt_adult_sel_mech = "stochastic";
static char *opt_adult_sel_proto = "full-gen-rep";
static double opt_crossover_prob = 0.0;
static size_t opt_children = 8;
static size_t opt_elites = 0;
static size_t opt_generations = 32;
static size_t opt_generation_print = 1;
static char *opt_log = NULL;
static double opt_mutation_prob = 0.0;
static size_t opt_nodes = 2;
static char *opt_parent_sel_mech = "fit-prop";
static char *opt_parent_sel_proto = "FIXME";
static size_t opt_crossover_points = 1;
static char *opt_population_log = NULL;
static char *opt_states = NULL;

static size_t opt_dft_winsize = 8;
static const char *opt_dft_winfunc = "rect";
static size_t opt_sim_statecount = 64;

static double (*dft_winfunc)(size_t, size_t);

static const struct option opt_long[] =
{
	{ "adults", 1, NULL, OPT_ADULTS },
	{ "adult-sel-mech", 1, NULL, OPT_ADULT_SEL_MECH },
	{ "adult-sel-proto", 1, NULL, OPT_ADULT_SEL_PROTO },
	{ "crossover-prob", 1, NULL, OPT_CROSSOVER_PROB },
	{ "children", 1, NULL, OPT_CHILDREN },
	{ "elites", 1, NULL, OPT_ELITES },
	{ "generation-print", 1, NULL, OPT_GENERATION_PRINT },
	{ "generations", 1, NULL, OPT_GENERATIONS },
	{ "help", 0, NULL, OPT_HELP },
	{ "log", 1, NULL, OPT_LOG },
	{ "mutation-prob", 1, NULL, OPT_MUTATION_PROB },
	{ "nodes", 1, NULL, OPT_NODES },
	{ "option", 1, NULL, OPT_OPTION },
	{ "parent-sel-mech", 1, NULL, OPT_PARENT_SEL_MECH },
	{ "parent-sel-proto", 1, NULL, OPT_PARENT_SEL_PROTO },
	{ "crossover-points", 1, NULL, OPT_CROSSOVER_POINTS },
	{ "population-log", 1, NULL, OPT_POPULATION_LOG },
	{ "states", 1, NULL, OPT_STATES },
	{ "version", 0, NULL, OPT_VERSION },
	{ NULL, 0, NULL, 0 },
};

static const char opt_short[] = ":a:C:c:e:G:g:hl:M:n:P:V";

static struct ddn_pt0_fitness_data ddnevo_pt0_fitness_data;
#define INIT_STATES 8
static mpz_t ddnevo_pt0_init_state[INIT_STATES];
static size_t ddnevo_pt0_init_state_count = INIT_STATES;

static struct ddn_pt0_fitness_op ddnevo_pt0_fitness_ops[] =
{
	{ .op1 = 0, .oper = DDN_PT0_FITNESS_OPERATOR_GT, .op2 = 1 },
	{ .op1 = 0, .oper = DDN_PT0_FITNESS_OPERATOR_GT, .op2 = 1 },
	{ .op1 = 0, .oper = DDN_PT0_FITNESS_OPERATOR_GT, .op2 = 1 },
	{ .op1 = 0, .oper = DDN_PT0_FITNESS_OPERATOR_GT, .op2 = 1 },
	{ .op1 = 0, .oper = DDN_PT0_FITNESS_OPERATOR_GT, .op2 = 1 },
	{ .op1 = 0, .oper = DDN_PT0_FITNESS_OPERATOR_GT, .op2 = 1 },
	{ .op1 = 0, .oper = DDN_PT0_FITNESS_OPERATOR_GT, .op2 = 1 },
	{ .op1 = 0, .oper = DDN_PT0_FITNESS_OPERATOR_GT, .op2 = 1 },
	//{ .op1 = 0, .oper = DDN_PT0_FITNESS_OPERATOR_SEP_BLOCK, .op2 = 1 },
};

static struct genotype *
ddnevo_gt0_create(void)
{
	return ddn_gt0_create(opt_nodes);
}

static struct phenotype *
ddnevo_pt0_create(const struct genotype *gt)
{
	return ddn_pt0_create(structof(gt, base, const struct ddn_gt0));
}

static int
ddnevo_pt0_test_fitness(struct phenotype *pt, const struct dtable *children)
{
	return ddn_pt0_fitness_test(structof(pt, base, struct ddn_pt0),
		&ddnevo_pt0_fitness_data, ddnevo_pt0_init_state, ddnevo_pt0_init_state_count,
		opt_sim_statecount, ddnevo_pt0_fitness_ops, ARRAY_SIZE(ddnevo_pt0_fitness_ops));
}

static const struct ga_ops ddnevo_t0_ops =
{
	.genotype_create = ddnevo_gt0_create,
	.phenotype_create = ddnevo_pt0_create,
	.phenotype_test_fitness = ddnevo_pt0_test_fitness,
};

struct option_cmp
{
	char *name;
	size_t len;
	union
	{
		size_t *z;
		const char **str;
	} u;
	int (*handle)(const struct option_cmp *o, const char *str);
};

static int
handle_strtoz(const struct option_cmp *o, const char *str)
{
	*(o->u.z) = atoi(str);
	return 0;
}

static int
handle_str(const struct option_cmp *o, const char *str)
{
	*(o->u.str) = str;
	return 0;
}

#define OPTION_CMP_DEFINE(s1, h) { .name = s1 "=", .len = sizeof(s1), .handle = h, }
#define OPTION_CMP_DEFINE_Z(s1, zinit) { .name = s1 "=", .len = sizeof(s1), .u.z = zinit, .handle = handle_strtoz, }
#define OPTION_CMP_DEFINE_STR(s1, strinit) { .name = s1 "=", .len = sizeof(s1), .u.str = strinit, .handle = handle_str, }
static const struct option_cmp option_cmp_lookup[] =
{
	OPTION_CMP_DEFINE_STR("dft-winfunc", &opt_dft_winfunc),
	OPTION_CMP_DEFINE_Z("dft-winsize", &opt_dft_winsize),
	OPTION_CMP_DEFINE_Z("sim-statecount", &opt_sim_statecount),
};

static int
parse_option(const char *str)
{
	for (size_t i = 0; i < ARRAY_SIZE(option_cmp_lookup); ++i)
	{
		const struct option_cmp *o = &option_cmp_lookup[i];
		if (!strncmp(o->name, str, o->len))
		{
			return o->handle(o, str + o->len);
		}
	}
	fprintf(stderr, "unknown option '%s'\n", str);
	return 1;
}

int
main(int argc, char **argv)
{
	int ret;
	struct ga ga;
	struct ga_settings ga_settings;
	struct ga_ops ga_ops;
	FILE *logfp = NULL, *poplogfp = NULL, *statesfp = NULL;
	struct uprand *upr, init_upr;

	prog = basename(argv[0]);

	while (1)
	{
		int opt;
		int ret;

		opt = getopt_long(argc, argv, opt_short, opt_long, NULL);
		if (opt == -1)
			break;
		switch (opt)
		{
		case OPT_ADULTS:
			opt_adults = atoi(optarg);
			break;
		case OPT_ADULT_SEL_MECH:
			opt_adult_sel_mech = optarg;
			break;
		case OPT_ADULT_SEL_PROTO:
			opt_adult_sel_proto = optarg;
			break;
		case OPT_CROSSOVER_PROB:
			opt_crossover_prob = atof(optarg);
			break;
		case OPT_CHILDREN:
			opt_children = atoi(optarg);
			break;
		case OPT_ELITES:
			opt_elites = atoi(optarg);
			break;
		case OPT_GENERATION_PRINT:
			opt_generation_print = atoi(optarg);
			break;
		case OPT_GENERATIONS:
			opt_generations = atoi(optarg);
			break;
		case OPT_HELP:
			usage();
			return 0;
		case OPT_LOG:
			opt_log = optarg;
			break;
		case OPT_MUTATION_PROB:
			opt_mutation_prob = atof(optarg);
			break;
		case OPT_NODES:
			opt_nodes = atoi(optarg);
			break;
		case OPT_OPTION:
			ret = parse_option(optarg);
			if (ret)
				return 1;
			break;
		case OPT_PARENT_SEL_MECH:
			opt_parent_sel_mech = optarg;
			break;
		case OPT_PARENT_SEL_PROTO:
			opt_parent_sel_proto = optarg;
			break;
		case OPT_CROSSOVER_POINTS:
			opt_crossover_points = atoi(optarg);
			break;
		case OPT_POPULATION_LOG:
			opt_population_log = optarg;
			break;
		case OPT_STATES:
			opt_states = optarg;
			break;
		case OPT_VERSION:
			version();
			return 0;
		case '?':
			fprintf(stderr, "%s: error: unknown option '%s'\n",
				prog, argv[optind - 1]);
			return 1;
		case ':':
			fprintf(stderr, "%s: error: missing argument\n", prog);
			return 2;
		default:
			fprintf(stderr, "%s: error: weird return value '%c'\n", prog, opt);
			return 3;
		}
	}

	dft_winfunc = dft_window_func_from_str(opt_dft_winfunc);
	if (!dft_winfunc)
	{
		fprintf(stderr, "%s: error: unknown window function '%s'\n",
			prog, opt_dft_winfunc);
		return 7;
	}

	ret = ddn_pt0_fitness_init(&ddnevo_pt0_fitness_data, opt_dft_winsize, dft_winfunc, opt_nodes);
	if (ret)
	{
		fprintf(stderr, "%s: error: failed to init fitness data: %s\n",
			prog, strerror(-ret));
		return 12;
	}

	ga_ops = ddnevo_t0_ops;

	if (!strcmp(opt_adult_sel_mech, "best-fit"))
	{
		ga_ops.adult_sel_mech = ga_adult_sel_mech_best_fitness;
	}
	else if (!strcmp(opt_adult_sel_mech, "stochastic"))
	{
		ga_ops.adult_sel_mech = ga_adult_sel_mech_stochastic;
	}
	else
	{
		fprintf(stderr, "%s: error: unknown adult selection mechanism '%s'\n",
			prog, opt_adult_sel_mech);
		return 20;
	}

	if (!strcmp(opt_adult_sel_proto, "full-gen-rep"))
	{
		ga_ops.adult_sel_proto = ga_adult_sel_proto_full_gen_rep;
	}
	else if (!strcmp(opt_adult_sel_proto, "gen-mix"))
	{
		ga_ops.adult_sel_proto = ga_adult_sel_proto_gen_mix;
	}
	else
	{
		fprintf(stderr, "%s: error: unknown adult selection protocol '%s'\n",
			prog, opt_adult_sel_proto);
		return 21;
	}

	if (!strcmp(opt_parent_sel_mech, "fit-prop"))
	{
		ga_ops.parent_sel_mech = fitness_proportionate_selection;
	}
	else if (!strcmp(opt_parent_sel_mech, "sigma-scale"))
	{
		ga_ops.parent_sel_mech = sigma_scale_selection;
	}
	else if (!strcmp(opt_parent_sel_mech, "tournament"))
	{
		//ga_ops.parent_sel_mech = tournament_selection;
		ga_ops.parent_sel_mech = NULL;
	}
	else
	{
		fprintf(stderr, "%s: error: unknown parent selection mechanism '%s'\n",
			prog, opt_parent_sel_mech);
		return 30;
	}

#if 0
	if (opt_adults > opt_children)
	{
		fprintf(stderr, "%s: error: number of adults '%zu' can not be larger than number of children '%zu'\n",
			prog, opt_adults, opt_children);
		return 14;
	}
#endif

	if (opt_elites > opt_adults)
	{
		fprintf(stderr, "%s: error: number of elites '%zu' can not be larger than the number of adults '%zu'\n",
			prog, opt_elites, opt_adults);
		return 15;
	}

	ga_settings.adults_count = opt_adults;
	ga_settings.children_count = opt_children;
	ga_settings.crossover_points = opt_crossover_points;
	ga_settings.crossover_prob = opt_crossover_prob;
	ga_settings.elites_count = opt_elites;
	ga_settings.generation_print = opt_generation_print;
	ga_settings.generation_count = opt_generations;
	ga_settings.mutation_prob = opt_mutation_prob;

	if (opt_log)
	{
		logfp = fopen(opt_log, "w");
		if (!logfp)
		{
			fprintf(stderr, "%s: error: failed to open file '%s': %s\n",
				prog, opt_log, strerror(errno));
			return 22;
		}
	}

	if (opt_population_log)
	{
		poplogfp = fopen(opt_population_log, "w");
		if (!poplogfp)
		{
			fprintf(stderr, "%s: error: failed to open file '%s': %s\n",
				prog, opt_population_log, strerror(errno));
			return 23;
		}
	}

	ga_settings.fitness_file = logfp;
	ga_settings.population_file = poplogfp;

	ret = ga_init(&ga, &ga_ops, &ga_settings);
	if (ret)
	{
		fprintf(stderr, "%s: error: failed to init ga: %s\n",
			prog, strerror(-ret));
		return 40;
	}

	ret = uprand_init(&init_upr);
	if (ret)
	{
		fprintf(stderr, "%s: error: failed to init pseudo random number generator: %s\n",
			prog, strerror(-ret));
		return 41;
	}
	for (size_t i = 0; i < ARRAY_SIZE(ddnevo_pt0_init_state); ++i)
	{
		unsigned int tmp;
		mpz_init(ddnevo_pt0_init_state[i]);
		tmp = prandi_range(&init_upr.base, 0, 1 << opt_nodes);
		mpz_set_ui(ddnevo_pt0_init_state[i], tmp);
	}
	//mpz_set_str(ddnevo_pt0_init_state[0], "00111100", 2);

	prand_cleanup(&init_upr.base);
#if 0
	mpz_set_str(ddnevo_pt0_init_state[0], "00000001", 2);
	mpz_set_str(ddnevo_pt0_init_state[1], "00000000", 2);
	mpz_set_str(ddnevo_pt0_init_state[2], "11111111", 2);
	mpz_set_str(ddnevo_pt0_init_state[3], "11000100", 2);
	mpz_set_str(ddnevo_pt0_init_state[4], "00010101", 2);
	mpz_set_str(ddnevo_pt0_init_state[5], "10001000", 2);
	mpz_set_str(ddnevo_pt0_init_state[6], "00111100", 2);
	mpz_set_str(ddnevo_pt0_init_state[7], "11000011", 2);
#endif
	upr = structof(ga.prand, base, struct uprand);

	fprintf(stdout, "adults: %zu\n", ga_settings.adults_count);
	fprintf(stdout, "children: %zu\n", ga_settings.children_count);
	fprintf(stdout, "crossover_points: %zu\n", ga_settings.crossover_points);
	fprintf(stdout, "crossover_prob: %f\n", ga_settings.crossover_prob);
	fprintf(stdout, "elites: %zu\n", ga_settings.elites_count);
	fprintf(stdout, "generation_print: %zu\n", ga_settings.generation_print);
	fprintf(stdout, "generations: %zu\n", ga_settings.generation_count);
	fprintf(stdout, "mutation_prob: %f\n", ga_settings.mutation_prob);
	fprintf(stdout, "log: %s\n", opt_log);
	fprintf(stdout, "population_log: %s\n", opt_population_log);
	fprintf(stdout, "adult_sel_mech: %s\n", opt_adult_sel_mech);
	fprintf(stdout, "adult_sel_proto: %s\n", opt_adult_sel_proto);
	fprintf(stdout, "parent_sel_mech: %s\n", opt_parent_sel_mech);
	fprintf(stdout, "seed: %d\n", upr->seed);
	fflush(stdout);

	if (opt_states)
	{
		statesfp = fopen(opt_states, "w");
		if (!statesfp)
		{
			fprintf(stderr, "%s: error: failed to open file '%s': %s\n",
				prog, opt_states, strerror(errno));
			return 50;
		}
		for (size_t i = 0; i < ARRAY_SIZE(ddnevo_pt0_init_state); ++i)
		{
			for (size_t j = opt_nodes; j > 0; --j)
			{
				fprintf(statesfp, "%d", mpz_tstbit(ddnevo_pt0_init_state[i], j - 1));
			}
			fprintf(statesfp, "\n");
		}
		fclose(statesfp);
	}

	ret = ga_run(&ga);
	if (ret)
	{
		fprintf(stderr, "%s: error: failed to run ga: %s\n",
			prog, strerror(-ret));
		return 41;
	}

	ga_cleanup(&ga);

	/* FIXME: run */

	if (poplogfp)
	{
		int tmp;

		do
		{
			tmp = fclose(poplogfp);
		} while (tmp == EINTR);

		if (tmp)
		{
			fprintf(stderr, "%s: error: failed to close file '%s': %s\n",
				prog, opt_population_log, strerror(errno));
			return 24;
		}
	}

	if (logfp)
	{
		int tmp;

		do
		{
			tmp = fclose(logfp);
		} while (tmp == EINTR);

		if (tmp)
		{
			fprintf(stderr, "%s: error: failed to close file '%s': %s\n",
				prog, opt_log, strerror(errno));
			return 24;
		}
	}

	if (ferror(stdin))
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
		return 10;
	}

	if (fflush(stdout))
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
		return 11;
	}

	if (ferror(stdout))
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
		return 12;
	}
	return 0;
}
