#include <dft.h>
#include <ddn/ddn_pt0.h>

#include <gmp.h>

#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>

#define OPT_EXT 0x7f

const char *prog;

static void
usage(void)
{
	fprintf(stderr, "usage: %s\n\n", prog);
	fprintf(stderr, "  -h, --help                   Show this help message.\n");
	fprintf(stderr, "  -n, --nodes <count>          Number of nodes in the network.\n");
	fprintf(stderr, "  -V, --version                Show version information.\n");
}

static void
version(void)
{
	fprintf(stderr, "ddnss v0\n");
}

static const struct option opt_long[] =
{
	{ "help", 0, NULL, 'h' },
	{ "nodes", 1, NULL, 'n' },
	{ "version", 0, NULL, 'V' },
	{ NULL, 0, NULL, 0 },
};

static const char opt_short[] = ":hn:V";

static size_t opt_nodes;

void
bindump(size_t val, size_t n, FILE *fp)
{
	for (size_t i = n; i > 0; --i)
	{
		if ((val & (1UL << (i - 1))))
			putc('1', fp);
		else
			putc('0', fp);
	}
}

int
main(int argc, char **argv)
{
	mpz_t znum;
	mpz_t znodes;
	void **statespace = NULL;
	void **prevstate = NULL;
	void *first = NULL;
	void *loopstate = NULL;

	prog = basename(argv[0]);

	while (1)
	{
		int opt;

		opt = getopt_long(argc, argv, opt_short, opt_long, NULL);
		if (opt == -1)
			break;
		switch (opt)
		{
		case 'h':
			usage();
			return 0;
		case 'n':
			opt_nodes = atoi(optarg);
			if (opt_nodes == 0)
			{
				fprintf(stderr, "%s: error: zero nodes is too small a network\n",
					prog);
				return 5;
			}
			break;
		case 'V':
			version();
			return 0;
		case '?':
			fprintf(stderr, "%s: error: unknown option\n", prog);
			return 1;
		case ':':
			fprintf(stderr, "%s: error: missing argument\n", prog);
			return 2;
		default:
			fprintf(stderr, "%s: error: weird return value '%c'\n", prog, opt);
			return 3;
		}
	}

	mpz_init(znum);
	mpz_init(znodes);
	mpz_set_ui(znodes, 1);
	mpz_mul_2exp(znodes, znodes, opt_nodes);
	while (1)
	{
		size_t count;
		int c;
		size_t idx;

		mpz_set_ui(znum, 0);
		do
		{
			c = fgetc(stdin);
		} while (isspace(c));

		if (c == EOF)
			break;
		for (count = 0; c != EOF && !isspace(c); c = fgetc(stdin))
		{
			mpz_mul_2exp(znum, znum, 1);
			if (c == '1' || c == '0')
			{
				if (c == '1')
					mpz_setbit(znum, 0);
				++count;
			}
			else
			{
				fprintf(stderr, "%s: error: garbage character '%c'\n", prog, c);
				return 12;
			}
		}

		if (!isspace(c) && c != EOF)
		{
			fprintf(stderr, "%s: error: garbage at the end of input\n", prog);
			return 13;
		}

		if (!opt_nodes)
		{
			opt_nodes = count;
			mpz_set_ui(znodes, 1);
			mpz_mul_2exp(znodes, znodes, opt_nodes);
		}

		if (count != opt_nodes)
		{
			fprintf(stderr, "%s: error: got '%zu' values, expected '%zu'\n",
				prog, count, opt_nodes);
			return 14;
		}

		if (!statespace)
		{
			statespace = calloc(1UL << opt_nodes, sizeof(*statespace));
			if (!statespace)
			{
				fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
				return 15;
			}
		}

		idx = mpz_get_ui(znum);
		if (idx >= (1UL << opt_nodes))
		{
			fprintf(stderr, "%s: error: state value '%zu' is larger than the number of nodes '%zu'\n",
				prog, idx, opt_nodes);
			return 16;
		}

		if (prevstate)
		{
			*prevstate = &statespace[idx];
			loopstate = prevstate;
			prevstate = &statespace[idx];
			if (*prevstate)
			{
				/* found cycle */
				break;
			}
		}
		else
		{
			first = &statespace[idx];
			prevstate = &statespace[idx];
		}

	}

	if (!statespace || !first || !prevstate || !loopstate)
	{
		fprintf(stderr, "%s: error: did not read a complete state\n",
			prog);
		return 20;
	}

	printf("digraph ss {\n");
	for (size_t i = 0; i < (1UL << opt_nodes); ++i)
	{
		if (statespace[i])
		{
			printf("\ts%zu [label=\"", i);
			bindump(i, opt_nodes, stdout);
			printf("\"];\n");
		}
	}

	void **p = first;
	void **pprev;
	do
	{
		size_t fidx = ((uintptr_t)p - (uintptr_t)statespace) / sizeof(uintptr_t);
		size_t tidx = ((uintptr_t)*p - (uintptr_t)statespace) / sizeof(uintptr_t);
		printf("\ts%zu -> s%zu;\n", fidx, tidx);
		pprev = p;
		p = *p;
	} while (pprev != loopstate);

	printf("}\n");

	if (ferror(stdin))
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
		return 10;
	}

	if (fflush(stdout))
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
		return 11;
	}

	if (ferror(stdout))
	{
		fprintf(stderr, "%s: error: %s\n", prog, strerror(errno));
		return 12;
	}

	return 0;
}
