/**
 * \ingroup dtable
 * \{
 * \file dtable.h
 * \brief Declarations for dynamic table.
 */

#ifndef DTABLE_H_
#define DTABLE_H_ 1

#ifndef DTABLE_DELTA
#define DTABLE_DELTA 16
#endif

#include <stdlib.h>

/**
 * \brief Dynamic table structure.
 */
struct dtable
{
	/**
	 * \brief Number of elements \a data can hold.
	 */
	size_t len;
	/**
	 * \brief Number of elements in \a data.
	 */
	size_t end;
	/**
	 * \brief Pointer to table of len elements of size \a size.
	 */
	void *data;
};


int dtable_cleanup(struct dtable *dt);

int _dtable_copy(struct dtable *dst, const struct dtable *src, size_t size);

/**
 * \brief Copy an entire dynamic table from \a src to \a dst.
 * \param dst Dynamic table to copy to.
 * \param src Dynamic table to copy from.
 * \param type Type of the elements.
 * \return 0 on success, -errno on failure.
 */
#define dtable_copy(dst, src, type) _dtable_copy(dst, src, sizeof(type))

#define DTABLE_INIT { .len = 0, .end = 0, .data = NULL }

int _dtable_init(struct dtable *dt, size_t len, size_t size);

/**
 * \brief Initialise \a dt to be a dynamic table of length \a len of elements
 * 	with a type of \a type.
 * \param dt Dynamic table to initialise.
 * \param len How many elements the table should initially have room for.
 * \param type Type of the elements.
 * \return 0 on success, -errno on failure.
 */
#define dtable_init(dt, len, type) _dtable_init(dt, len, sizeof(type))

int _dtable_insert(struct dtable *dt, const void *data, size_t index,
	size_t size);

/**
 * \brief Insert \a data into \a dt as element \a index.
 * \param dt Dynamic table to insert into.
 * \param data Pointer to element to insert into table.
 * \param index Index of element to place \a data as.
 * \param type Type of the elements.
 * \return 0 on success, -errno on failure.
 */
#define dtable_insert(dt, data, index, type) \
	_dtable_insert(dt, data, index, sizeof(type))

void _dtable_remove(struct dtable *dt, size_t index, size_t size);

/**
 * \brief Remove element at \a index from \a dt.
 * \param dt Dynamic table to remove from.
 * \param index Index of element to remove.
 * \param type Type of the elements.
 */
#define dtable_remove(dt, index, type) \
	_dtable_remove(dt, index, sizeof(type))

/**
 * \brief Remove all elements from dynamic table \a dt.
 * \param dt Dynamic table to remove all elements from.
 */
static inline void
dtable_wipe(struct dtable *dt)
{
	dt->end = 0;
}

/**
 * \brief Get pointer to element \a index of the dynamic table of type \a type.
 * \param dt Pointer to the dynamic table.
 * \param index Index of the element to get a pointer to.
 * \param type The type of the elements in the table.
 * \return A pointer to element \a index of the table.
 */
#define DTABLE_DATA(dt, index, type) (((type *)((dt)->data))[index])

/**
 * \brief Get pointer to element \a index of the dynamic table.
 * \param dt Pointer to the dynamic table.
 * \param index Index of the element to get a pointer to.
 * \param size Size of an element.
 * \return A pointer to element \a index of the table.
 */
static inline void *
_dtable_ptr(struct dtable *dt, size_t index, size_t size)
{
	return ((char *)dt->data) + (index * size);
}

/**
 * \brief Get a pointer located in element \a index of the dynamic table.
 * \param dt Pointer to the dynamic table.
 * \param index Index of the element that contains a pointer.
 * \param size Size of an element.
 * \return A pointer the is located in element \a index of the table.
 */
static inline void *
_dtable_dataptr(struct dtable *dt, size_t index, size_t size)
{
	return *(void **)_dtable_ptr(dt, index, size);
}

/* FIXME: */
static inline void *
_dtable_data(struct dtable *dt, size_t index, size_t size)
{
	return _dtable_ptr(dt, index, size);
}

/**
 * \brief Iterate over all elements in \a dt.
 * \brief elem Pointer to hold the current element.
 * \brief dt The dynamic table to iterate over.
 * \brief type The type of the elements of the table.
 */
#define dtable_for_each(elem, dt, type) \
	for ((elem) = &DTABLE_DATA((dt), 0, (type)); \
		(elem) < &DTABLE_DATA((dt), (dt)->end, (type)); \
		++(elem))

static inline size_t
dtable_nnext(const struct dtable *dt, size_t idx, size_t n)
{
	return (idx + n) % dt->end;
}

static inline size_t
dtable_next(const struct dtable *dt, size_t idx)
{
	return dtable_nnext(dt, idx, 1);
}

static inline size_t
dtable_nprev(const struct dtable *dt, size_t idx, size_t n)
{
	if (n > idx)
		return dt->end - (n - idx);
	return idx - n;
}

static inline size_t
dtable_prev(const struct dtable *dt, size_t idx)
{
	return dtable_nprev(dt, idx, 1);
}

#endif /* !DTABLE_H_ */
/**
 * \}
 */
