#include <dbg.h>
#include <util.h>
#include <ddn/ddn_gt0.h>
#include <ddn/ddn.h>
#include <ddn/ddn_pt0.h>
#include <prand.h>

#include <ga/phenotype.h>

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <stdint.h>
#include <errno.h>
#include <ctype.h>

static int ddn_gt0_destroy(struct genotype *gt);
static int ddn_gt0_mutate(struct genotype *geno, struct prand *prand, size_t feature);
static int ddn_gt0_crossover(struct genotype *child1, struct genotype *child2, size_t n, struct prand *prand);
static int ddn_gt0_init_random(struct genotype *gt, struct prand *prand);
static void ddn_gt0_fprint(const struct genotype *gt, FILE *out);
static struct genotype *ddn_gt0_clone(const struct genotype *gt);

const struct genotype_ops
ddn_gt0_ops =
{
	.mutate = ddn_gt0_mutate,
	.crossover = ddn_gt0_crossover,
	.init_random = ddn_gt0_init_random,
	.fprint = ddn_gt0_fprint,
	.clone = ddn_gt0_clone,
};

struct genotype *
ddn_gt0_create(size_t n)
{
	struct ddn_gt0 *ret;

	ret = malloc(sizeof(*ret) + bit_count(ddn_gt0_bits(n)) * sizeof(ret->bits[0]));
	if (!ret)
	{
		int tmp = errno;
		dbg_error("%s\n", strerror(tmp));
		errno = tmp;
		return NULL;
	}
	ret->count = n;
	ret->base.features = 2 * n;
	ret->base.cleanup = ddn_gt0_destroy;
	ret->base.ops = &ddn_gt0_ops;

	return &ret->base;
}

static int
ddn_gt0_destroy(struct genotype *gt)
{
	struct ddn_gt0 *gt0;

	gt0 = structof(gt, base, struct ddn_gt0);
	free(gt0);
	return 0;
}

size_t
ddn_gt0_param2num(const unsigned long *bits, size_t offset, size_t len)
{
	size_t num;

	num = 0;
	/* We interpret param as big endian */
	for (size_t i = 0; i < len; ++i)
	{
		if (bit_isset(bits, offset + i))
			num = (num << 1) | 1;
		else
			num = num << 1;
	}

	return num;
}

void
ddn_gt0_num2param(unsigned long *bits, size_t offset, size_t len, size_t num)
{
	for (size_t i = 0; i < len; ++i)
	{
		if ((num & 1))
			bit_set(bits, offset + i);
		else
			bit_clr(bits, offset + i);
	}
}

static int
ddn_gt0_mutate(struct genotype *gt, struct prand* prand, size_t feature)
{
	struct ddn_gt0 *gt0 = structof(gt, base, struct ddn_gt0);
	size_t param_bits;
	size_t bit_base;

	dbg_assert(feature < gt0->base.features);
	param_bits = ilog2z(gt0->count);
	bit_base = (gt0->count + param_bits) * (feature / 2);

	if ((feature & 1))
	{
		/* we should mutate connection of feature */
		int pos;
		pos = prandi_range(prand, 0, gt0->count);
		bit_flip(gt0->bits, bit_base + param_bits + pos);
	}
	else
	{
		/* we should mutate parameter of feature */
		int inc;
		size_t num;

		/* FIXME: wrap around handling? */
		num = ddn_gt0_param2num(gt0->bits, bit_base, param_bits);
		inc = prandi_range(prand, 0, 2);
		if (inc)
		{
			++num;
		}
		else
		{
			--num;
		}
		ddn_gt0_num2param(gt0->bits, bit_base, param_bits, num);
	}
	return 0;
}

static int
size_t_cmp(const void *p1, const void *p2)
{
	size_t v1, v2;

	v1 = *(const size_t *)p1;
	v2 = *(const size_t *)p2;

	if (v1 < v2)
		return -1;
	if (v1 > v2)
		return 1;
	return 0;
}

static int
ddn_gt0_crossover(struct genotype *out1, struct genotype *out2, size_t n,
		struct prand *prand)
{
	struct ddn_gt0 *child1 = structof(out1, base, struct ddn_gt0);
	struct ddn_gt0 *child2 = structof(out2, base, struct ddn_gt0);
	/* FIXME: dynami n-point crossover */
	size_t points[n];
	size_t param_bits;

	assert(child1->count == child2->count);
	assert(n < child1->count * 2);
	for (size_t j = 0; j < n; ++j)
	{
		points[j] = prandi_range(prand, 0, child1->count * 2);
	}
	qsort(points, n, sizeof(points[0]), size_t_cmp);
	param_bits = ilog2z(child1->count);

	for (size_t i = 0; i < n; ++i)
	{
		size_t bit_base;
		size_t count;
	
		bit_base = (child1->count + param_bits) * (points[i] / 2);
		if ((points[i] & 1))
		{
			count = child1->count;
			bit_base += param_bits;
		}
		else
		{
			count = param_bits;
		}

		for (size_t j = 0; j < count; ++j)
		{
			int set1, set2;

			set1 = bit_isset(child1->bits, bit_base + j);
			set2 = bit_isset(child2->bits, bit_base + j);

			if (set1)
				bit_set(child2->bits, bit_base + j);
			else
				bit_clr(child2->bits, bit_base + j);

			if (set2)
				bit_set(child1->bits, bit_base + j);
			else
				bit_clr(child1->bits, bit_base + j);
		}
	}

	return 0;
}

static int
ddn_gt0_init_random(struct genotype *gt, struct prand *prand)
{
	struct ddn_gt0 *gt0 = structof(gt, base, struct ddn_gt0);
	size_t r, bits;

	bits = ddn_gt0_bits(gt0->count);
	memset(gt0->bits, 0, bit_count(bits) * sizeof(gt0->bits[0]));
	r = prandi_range(prand, 0, bits);
	while (r > 0)
	{
		size_t i;

		i = prandi_range(prand, 0, ddn_gt0_bits(gt0->count));
		while (bit_isset(gt0->bits, i))
		{
			i = (i + 1) % bits;
		}
		bit_set(gt0->bits, i);
		--r;
	}

	return 0;
}

static void
ddn_gt0_fprint(const struct genotype *gt, FILE *out)
{
	const struct ddn_gt0 *gt0 = structof(gt, base, const struct ddn_gt0);
	size_t len;

	fprintf(out, "%zu ", gt0->count);
	len = ddn_gt0_bits(gt0->count);
	for (size_t i = 0; i < len; ++i)
	{
		if (bit_isset(gt0->bits, i))
			fputc('1', out);
		else
			fputc('0', out);
	}
	//fputc('\n', out);
}

struct genotype *
ddn_gt0_fscan(FILE *fp)
{
	char buf[24];
	size_t len;
	struct genotype *gt;
	size_t n;
	int c;
	struct ddn_gt0 *gt0;
	size_t bits, count;

	gt = NULL;
	do
	{
		c = fgetc(fp);
		if (c == -1)
		{
			int tmp = errno;
			dbg_error("%s\n", strerror(errno));
			errno = tmp;
			goto err_out;
		}
	} while (isspace(c));

	for (len = 0; len < sizeof(buf) - 1 && !isspace(c) && c != EOF; ++len, c = fgetc(fp))
	{
		buf[len] = c;
	}

	if (c == EOF)
	{
		int tmp = errno;
		dbg_error("%s\n", strerror(errno));
		errno = tmp;
		goto err_out;
	}
	else if (!isspace(c))
	{
		dbg_error("parameter contains too many characters\n");
		goto err_out;
	}

	buf[len] = 0;

	n = strtoul(buf, NULL, 0);
	if (n == ULONG_MAX)
	{
		int tmp = errno;
		dbg_error("%s\n", strerror(errno));
		errno = tmp;
		goto err_out;
	}

	gt = ddn_gt0_create(n);
	if (!gt)
	{
		int tmp = errno;
		dbg_error("%s\n", strerror(errno));
		errno = tmp;
		goto err_out;
	}

	gt0 = structof(gt, base, struct ddn_gt0);
	bits = ddn_gt0_bits(gt0->count);

	do
	{
		c = fgetc(fp);
		if (c == -1)
		{
			dbg_error("%s\n", strerror(errno));
			goto bits_fail_out;
		}
	} while (isspace(c));

	count = 0;
	for (count = 0; count < bits && !isspace(c) && c != EOF; ++count, c = fgetc(fp))
	{
		if (c == '0')
		{
			bit_clr(gt0->bits, count);
		}
		else if (c == '1')
		{
			bit_set(gt0->bits, count);
		}
		else
		{
			dbg_error("invalid character '%c'\n", c);
			goto bits_fail_out;
		}
	}

	if (count != bits)
	{
		dbg_error("got %zu bits, expected %zu\n", count, bits);
		goto bits_fail_out;
	}
	else if (c != EOF && !isspace(c))
	{
		dbg_error("garbage at end of input\n");
		goto bits_fail_out;
	}
	else if (c == EOF && !feof(fp))
	{
		int tmp = errno;
		dbg_error("%s\n", strerror(errno));
		errno = tmp;
		goto bits_fail_out;
	}

	return gt;

bits_fail_out:
	ddn_gt0_destroy(gt);
err_out:
	return NULL;
}

static struct genotype *
ddn_gt0_clone(const struct genotype *gt)
{
	const struct ddn_gt0 *gt0 = structof(gt, base, const struct ddn_gt0);
	struct ddn_gt0 *tmp;
	struct genotype *ret;

	ret = ddn_gt0_create(gt0->count);
	if (!ret)
	{
		return NULL;
	}
	tmp = structof(ret, base, struct ddn_gt0);
	ret->ops = gt->ops;
	ret->cleanup = ddn_gt0_destroy;

	memcpy(tmp->bits, gt0->bits, bit_count(ddn_gt0_bits(gt0->count)) * sizeof(gt0->bits[0]));
	return ret;
}

#if 0
struct phenotype *
ddn_gt0_develop(const struct genotype *gt)
{
	struct ddn_gt0 *gt0 = structof(gt, base, struct ddn_gt0);
	return ddn_pt0_create(gt0->count, gt);
}
#endif
