#include <ddn/ddn.h>
#include <ddn/ddn_pt0.h>
#include <ddn/ddn_gt0.h>
#include <util.h>
#include <dft.h>

#include <complex.h>
#include <math.h>
#include <gmp.h>

#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

struct ddn_pt0_state_functor
{
	struct ddn_state_functor base;
	size_t param;
};

static ddn_state_t
ddn_pt0_state_func(const struct ddn_state_functor *base, const struct ddn_node **input, size_t len)
{
	const struct ddn_pt0_state_functor *fn = structof(base, base, const struct ddn_pt0_state_functor);
	ddn_state_t state;
	uintptr_t sum;

	sum = 0;
	for (size_t i = 0; i < len; ++i)
	{
		if (input[i])
			sum += input[i]->state.u.num;
	}
	state.u.num = (sum > fn->param ? 1 : 0);

	return state;
}

int
ddn_pt0_state_func_fprint(const struct ddn_state_functor *dsf, size_t n, FILE *fp)
{
	const struct ddn_pt0_state_functor *fun = structof(dsf, base,
		const struct ddn_pt0_state_functor);

	return fprintf(fp, "sum(i[%zu:0]) > %zu", n - 1, fun->param);
}

static struct ddn_pt0_state_functor *
ddn_pt0_state_functor_create(size_t param)
{
	struct ddn_pt0_state_functor *fn;

	fn = malloc(sizeof(*fn));
	if (!fn)
	{
		return fn;
	}

	fn->param = param;
	fn->base.state_next = ddn_pt0_state_func;
	fn->base.func_fprint = ddn_pt0_state_func_fprint;
	return fn;
}

struct phenotype *
ddn_pt0_create(const struct ddn_gt0 *gt0)
{
	size_t param_bits;
	struct ddn_pt0 *pt0;
	size_t i;
	int ret;

	pt0 = malloc(sizeof(*pt0));
	if (!pt0)
	{
		return 0;
	}

	ret = ddn_init(&pt0->ddn, gt0->count);
	if (ret)
	{
		free(pt0);
		pt0 = NULL;
	}

	assert(gt0->count == pt0->ddn.n);
	param_bits = ilog2z(gt0->count);

	for (i = 0; i < gt0->count; ++i)
	{
		size_t bits_base;
		struct ddn_pt0_state_functor *fn;
		size_t param;

		bits_base = (param_bits + gt0->count) * i;
		param = ddn_gt0_param2num(gt0->bits, bits_base, param_bits);
		fn = ddn_pt0_state_functor_create(param);
		if (!fn)
		{
			goto error_out;
		}
		pt0->ddn.node[i].state_next = &fn->base;
		for (size_t j = 0; j < gt0->count; ++j)
		{
			if (bit_isset(gt0->bits, bits_base + param_bits + j))
			{
				pt0->ddn.node[i].input[j] = &pt0->ddn.node[j];
				pt0->ddn.node[j].output[i] = &pt0->ddn.node[i];
			}
		}
	}
	pt0->base.cleanup = ddn_pt0_destroy;

	return &pt0->base;

error_out:
	do
	{
		--i;
		free(pt0->ddn.node[i].state_next);
	} while(i != 0);
	ddn_pt0_destroy(&pt0->base);

	return NULL;
}

int
ddn_pt0_fitness_init(struct ddn_pt0_fitness_data *fit, size_t window_size,
		double (*window_func)(size_t, size_t), size_t nodes)
{
	int ret;

	dbg_assert(window_size >= 2);
	mpz_init(fit->znum);
	mpz_init(fit->znum_max);
	mpf_init(fit->fnum);
	mpf_init(fit->fnum_max);

	ret = dft_plan_init(&fit->dft, window_size, window_func);
	if (ret)
	{
		dbg_error("%s\n", strerror(-ret));
		goto dft_plan_fail;
	}

	fit->sorted_result_max = malloc(fit->dft.transformed_size *
			sizeof(*(fit->sorted_result_max)));
	if (!fit->sorted_result_max)
	{
		ret = -errno;
		dbg_error("%s\n", strerror(-ret));
		goto sorted_result_max_fail;
	}

	fit->sorted_result_min = malloc(fit->dft.transformed_size *
			sizeof(*(fit->sorted_result_min)));
	if (!fit->sorted_result_min)
	{
		ret = -errno;
		dbg_error("%s\n", strerror(-ret));
		goto sorted_result_min_fail;
	}

	mpz_init(fit->coverage);
	mpf_init(fit->fcoverage_max);
	mpf_set_ui(fit->fcoverage_max, 1);
	mpf_mul_2exp(fit->fcoverage_max, fit->fcoverage_max, nodes);
	mpf_sub_ui(fit->fcoverage_max, fit->fcoverage_max, 1);
	mpf_init(fit->fcoverage);

	return 0;

	free(fit->sorted_result_min);
sorted_result_min_fail:
	free(fit->sorted_result_max);
sorted_result_max_fail:
	dft_plan_cleanup(&fit->dft);
dft_plan_fail:
	mpf_clear(fit->fnum_max);
	mpf_clear(fit->fnum);
	mpz_clear(fit->znum_max);
	mpz_clear(fit->znum);
	return ret;
}

double
ddn_pt0_fitness_sample(struct ddn_pt0_fitness_data *fit, size_t bits)
{
	double sample;

	mpf_set_z(fit->fnum, fit->znum);

	mpz_set_ui(fit->znum_max, 1);
	mpz_mul_2exp(fit->znum_max, fit->znum_max, bits);
	mpf_set_z(fit->fnum_max, fit->znum_max);
	mpf_sub_ui(fit->fnum_max, fit->fnum_max, 1);
	mpf_div_ui(fit->fnum_max, fit->fnum_max, 2);
	mpf_sub(fit->fnum, fit->fnum, fit->fnum_max);
	mpf_div(fit->fnum, fit->fnum, fit->fnum_max);
	sample = mpf_get_d(fit->fnum);
	return sample;
}

static void
ddn_pt0_fitness_test_dft(struct ddn *ddn, struct ddn_pt0_fitness_data *fit,
	mpz_t init_state, size_t state_count)
{
	size_t i;
	int repeat;

	dft_init(&fit->dft);
	mpz_set_ui(fit->coverage, 0);
	ddn_state_fromz(ddn, init_state);
	repeat = 0;
	i = 0;
	while (1)
	{
		double sample;
		unsigned long num;

		ddn_state_toz(ddn, fit->znum);
		num = mpz_get_ui(fit->znum);
		if (!repeat && mpz_tstbit(fit->coverage, num))
		{
			repeat = 1;
			mpz_set_ui(fit->coverage, 0);
		}
		mpz_setbit(fit->coverage, num);
		sample = ddn_pt0_fitness_sample(fit, ddn->n);
		dft_sample_add(&fit->dft, sample);
		if (i >= state_count)
			break;

		ddn_step(ddn);
		++i;
	}
	dft_finish(&fit->dft);
}

static double
ddn_pt0_fitness_calc(const struct ddn_pt0_fitness_result * const res_max, size_t res_max_len,
		const struct ddn_pt0_fitness_result * const res_min, size_t res_min_len,
		const struct ddn_pt0_fitness_op * const op, double cov)
{
	/* FIXME: implement calculation of CA ops */
	double fitness;
	const struct ddn_pt0_fitness_result *res1, *res2;

	fitness = 0.0;

	if (op->op1 >= res_max_len || op->op2 >= res_max_len)
	{
		double tmp;
		if (cov != 0.0)
			tmp = log(cov) - log(1.0 / 256.0);
		else
			tmp = 0.0;

		return tmp; // * 100.0;
	}

	res1 = &res_max[op->op1];
	res2 = &res_max[op->op2];

#if 0
	if (op[i].op1 < res_max_len)
	{
		res1 = &res_max[op[i].op1];
	}
	else
	{
		return cov * 100.0;
	}

	if (op[i].op2 < res_max_len)
	{
		if (*(res_max[op[i].op2].val) > *(res_min[0].val))
			res2 = &res_max[op[i].op2];
		else if (res_min_len > 1)
			res2 = &res_min[0];
		else
			return cov * 100.0;
	}
	else if (res_min_len > 1)
	{
		res2 = &res_min[0];
	}
	else
	{
		return cov * 100.0;
	}
#endif

	switch (op->oper)
	{
	case DDN_PT0_FITNESS_OPERATOR_GT:
		fitness += (*(res1->val) - *(res2->val)) * (double)op->op2;
		break;
	case DDN_PT0_FITNESS_OPERATOR_LT:
		fitness += (*(res2->val) - *(res1->val)) * (double)op->op2;
		break;
#if 0
	case DDN_PT0_FITNESS_OPERATOR_SEP_BLOCK:
		if ((res[op[i].op1].idx == res[op[i].op2].idx - 1) ||
				(res[op[i].op2].idx == res[op[i].op1].idx - 1))
		{
			fitness -= (*(res[op[i].op1].val));
		}
		break;
#endif
	default:
		dbg_error("unknown op\n");
	}
	//dbg_assert(fitness != 0.0);

	return fitness;
}

static int
ddn_pt0_fitness_test_cmp(const void *p1, const void *p2)
{
	const struct ddn_pt0_fitness_result *op1, *op2;

	op1 = (struct ddn_pt0_fitness_result const *)p1;
	op2 = (struct ddn_pt0_fitness_result const *)p2;

	if (*(op1->val) > *(op2->val))
		return -1;
	else if (*(op1->val) < *(op2->val))
		return 1;
	return 0;
}

int
ddn_pt0_fitness_test(struct ddn_pt0 *pt0, struct ddn_pt0_fitness_data *fit,
	mpz_t *init_state, size_t init_state_count, size_t state_count,
	const struct ddn_pt0_fitness_op * const op, size_t op_count)
{
	double fitness;
	int ret;

	dbg_assert(fit->dft.window_size >= 2);
	dbg_assert(op_count);
	dbg_assert(init_state_count);
	dbg_assert(op_count == init_state_count);
#if 0
	fitness = calloc(init_state_count, sizeof(*fitness));
	if (!fitness)
	{
		ret = -errno;
		dbg_error("%s\n", strerror(-ret));
		return ret;
	}
#endif
	fitness = 0.0;
	for (size_t i = 0; i < init_state_count; ++i)
	{
		int max_point;
		size_t max_pos, min_pos, j;
		unsigned long popcount;
		double dft_fit, cov_fit;

		ddn_pt0_fitness_test_dft(&pt0->ddn, fit, init_state[i], state_count - 1);
		dbg_assert(fit->dft.transformed_size >= 3);

		popcount = mpz_popcount(fit->coverage);
		mpf_set_ui(fit->fcoverage, popcount);
		mpf_div(fit->fcoverage, fit->fcoverage, fit->fcoverage_max);
		cov_fit = mpf_get_d(fit->fcoverage);

		fit->sorted_result_max[0].idx = 1;
		fit->sorted_result_max[0].val = &fit->dft.result[1];
		/* start with max_point find, since if all freqs are zero, we add atleast a last result */
		max_point = 1;
		for (j = 2; j < fit->dft.transformed_size; ++j)
		{
			if (fit->dft.result[j - 1] > fit->dft.result[j])
			{
				max_point = 0;
				break;
			}
			else if (fit->dft.result[j - 1] < fit->dft.result[j])
			{
				max_point = 1;
				break;
			}
		}
		/* FIXME: what do we do with DC component */
		for (max_pos = 1, min_pos = 0; j < fit->dft.transformed_size; ++j)
		{
			if (max_point)
			{
				if (fit->dft.result[j] < fit->dft.result[j - 1])
				{
					fit->sorted_result_max[max_pos].idx = j - 1;
					fit->sorted_result_max[max_pos].val = &fit->dft.result[j - 1];
					max_point = 0;
					++max_pos;
				}
			}
			else
			{
				if (fit->dft.result[j] > fit->dft.result[j - 1])
				{
					fit->sorted_result_min[min_pos].idx = j - 1;
					fit->sorted_result_min[min_pos].val = &fit->dft.result[j - 1];
					max_point = 1;
					++min_pos;
				}
			}
		}

		if (max_point)
		{
			fit->sorted_result_max[max_pos].idx = j - 1;
			fit->sorted_result_max[max_pos].val = &fit->dft.result[j - 1];
			++max_pos;
		}
		else
		{
			fit->sorted_result_min[min_pos].idx = j - 1;
			fit->sorted_result_min[min_pos].val = &fit->dft.result[j - 1];
			++min_pos;
		}

		qsort(fit->sorted_result_max, max_pos,
			sizeof(*(fit->sorted_result_max)), ddn_pt0_fitness_test_cmp);
		qsort(fit->sorted_result_min, min_pos,
			sizeof(*(fit->sorted_result_min)), ddn_pt0_fitness_test_cmp);
		fit->sorted_result_min[min_pos].idx = 0;
		fit->sorted_result_min[min_pos].val = &fit->dft.result[0];
		++min_pos;

		dft_fit = ddn_pt0_fitness_calc(fit->sorted_result_max, max_pos,
				fit->sorted_result_min, min_pos,
				&op[i], cov_fit); //fabs(fit->dft.result[0]));
		//dft_fit = dft_fit / ((double)state_count);
		fitness += (dft_fit); // + cov_fit * 0.5);
		//if (dft_fit < 10.0)
		//	break;

		//fitness_acc += ddn_pt0_fitness_test_find_diff(fit);
	}

	pt0->base.fitness = fitness;
	return 0;
}

int
ddn_pt0_fitness_cleanup(struct ddn_pt0_fitness_data *fit)
{
	mpf_clear(fit->fcoverage);
	mpf_clear(fit->fcoverage_max);
	mpz_clear(fit->coverage);
	free(fit->sorted_result_min);
	free(fit->sorted_result_max);
	dft_plan_cleanup(&fit->dft);
	mpf_clear(fit->fnum_max);
	mpf_clear(fit->fnum);
	mpz_clear(fit->znum_max);
	mpz_clear(fit->znum);
	return 0;
}

int
ddn_pt0_destroy(struct phenotype *pt)
{
	struct ddn_pt0 *pt0 = structof(pt, base, struct ddn_pt0);

	for (size_t i = 0; i < pt0->ddn.n; ++i)
	{
		free(pt0->ddn.node[i].state_next);
		pt0->ddn.node[i].state_next = NULL;
	}

	ddn_cleanup(&pt0->ddn);
	free(pt0);
	return 0;
}

struct ddn *
ddn_pt0_ddn(struct phenotype *pt)
{
	struct ddn_pt0 *pt0 = structof(pt, base, struct ddn_pt0);

	return &pt0->ddn;
}
