#include <dtable.h>
#include <dbg.h>
#include <util.h>
#include <ga/individual.h>
#include <ga/ga.h>
#include <prand.h>
#include <uprand.h>

#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <math.h>

int
ga_init(struct ga *ga, struct ga_ops *ops, struct ga_settings *settings)
{
	int ret;

	ga->ops = ops;
	ga->settings = settings;
	ga->mutation_possi = 0;
	ga->mutation_count = 0;
	ga->crossover_possi = 0;
	ga->crossover_count = 0;
	ga->best = NULL;

	/* FIXME: init ga->prand */

	ret = dtable_init(&ga->adults, GA_DTABLE_SIZE, struct individual);
	if (ret)
	{
		dbg_error("%s\n", strerror(-ret));
		goto init1_fail;
	}
	ret = dtable_init(&ga->children, GA_DTABLE_SIZE, struct individual);
	if (ret)
	{
		dbg_error("%s\n", strerror(-ret));
		goto init2_fail;
	}

	ret = dtable_init(&ga->parents, GA_DTABLE_SIZE, struct individual);
	if (ret)
	{
		dbg_error("%s\n", strerror(-ret));
		goto init3_fail;
	}

	ga->prand = uprand_create();
	if (!ga->prand)
	{
		ret = -errno;
		dbg_error("%s\n", strerror(-ret));
		goto uprand_create_fail;
	}

	for (size_t i = 0; i < ga->settings->children_count; ++i)
	{
		struct individual ind;
		struct genotype *geno;

		geno = ga->ops->genotype_create();
		if (!geno)
		{
			ret = -errno;
			dbg_error("%s\n", strerror(-ret));
			goto individual_fail;
		}

		ret = individual_init(&ind, geno);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			genotype_cleanup(geno);
			goto individual_fail;
		}

		ret = genotype_init_random(ind.genotype, ga->prand);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			individual_cleanup(&ind);
			goto individual_fail;
		}

		ret = dtable_insert(&ga->children, &ind, ga->children.end, struct individual);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			individual_cleanup(&ind);
			goto individual_fail;
		}
	}

	return 0;

individual_fail:
	for (size_t i = 0; i < ga->children.end; ++i)
	{
		individual_cleanup(&DTABLE_DATA(&ga->children, i, struct individual));
	}
	prand_cleanup(ga->prand);
uprand_create_fail:
	dtable_cleanup(&ga->parents);
init3_fail:
	dtable_cleanup(&ga->children);
init2_fail:
	dtable_cleanup(&ga->adults);
init1_fail:
	return ret;
}

int
ga_children_develop(struct ga *ga)
{
	int ret = 0;

	for (size_t i = 0; i < ga->children.end; ++i)
	{
		struct individual *p;
		struct phenotype *pheno;

		p = &DTABLE_DATA(&ga->children, i, struct individual);
		pheno = ga->ops->phenotype_create(p->genotype);
		if (!pheno)
		{
			ret = -errno;
			dbg_error("%s\n", strerror(-ret));
			break;
		}
		p->phenotype = pheno;
	}

	return ret;
}

int
ga_cleanup(struct ga *ga)
{
	prand_cleanup(ga->prand);
	for (size_t i = 0; i < ga->adults.end; ++i)
	{
		struct individual *p;

		p = &DTABLE_DATA(&ga->adults, i, struct individual);
		individual_cleanup(p);
	}
	dtable_cleanup(&ga->adults);
	for (size_t i = 0; i < ga->children.end; ++i)
	{
		struct individual *p;

		p = &DTABLE_DATA(&ga->children, i, struct individual);
		individual_cleanup(p);
	}
	dtable_cleanup(&ga->children);
	for (size_t i = 0; i < ga->parents.end; ++i)
	{
		struct individual *p;

		p = &DTABLE_DATA(&ga->parents, i, struct individual);
		individual_cleanup(p);
	}
	dtable_cleanup(&ga->parents);
	return 0;
}

int
ga_children_test_fitness(struct ga *ga)
{
	for (size_t i = 0; i < ga->children.end; ++i)
	{
		struct individual *p;

		p = &DTABLE_DATA(&ga->children, i, struct individual);
		ga->ops->phenotype_test_fitness(p->phenotype, &ga->children);
	}

	return 0;
}

int
ga_adult_selection(struct ga *ga)
{
	int ret;
	double avg, sum;
	double stddev, acc;
	double best;
	struct individual *indbest = NULL;

	ret = ga->ops->adult_sel_proto(ga);
	if (ret)
	{
		dbg_error("%s\n", strerror(-ret));
		return ret;
	}

	ret = ga->ops->adult_sel_mech(ga);
	if (ret)
	{
		dbg_error("%s\n", strerror(-ret));
		return ret;
	}

	sum = 0.0;
	best = 0.0;
	/* average */
	for (size_t i = 0; i < ga->adults.end; ++i)
	{
		struct individual *p;

		p = &DTABLE_DATA(&ga->adults, i, struct individual);
		sum += p->phenotype->fitness;
		if (i == 0 || p->phenotype->fitness > best)
		{
			indbest = p;
			best = p->phenotype->fitness;
		}
	}
	avg = sum / (double)ga->adults.end;

	acc = 0.0;
	/* standard deviation */
	for (size_t i = 0; i < ga->adults.end; ++i)
	{
		struct individual *p;
		double tmp;

		p = &DTABLE_DATA(&ga->adults, i, struct individual);
		tmp = p->phenotype->fitness - avg;
		acc += tmp * tmp;
	}
	stddev = sqrt(acc / (double)ga->adults.end);

	dbg_assert(indbest != NULL);
	dbg_assert(ga->adults.end != 0);
	ga->fitness_stddev = stddev;
	ga->fitness_avg = avg;
	ga->fitness_best = best;
	ga->best = indbest;

	return 0;
}

void
ga_population_fprint(struct ga *ga, FILE *out)
{
	if (!(ga->generation % ga->settings->generation_print))
	{
		for (size_t i = 0; i < ga->adults.end; ++i)
		{
			struct individual *p;

			p = &DTABLE_DATA(&ga->adults, i, struct individual);
			fprintf(out, "%zu %f ", ga->generation, p->phenotype->fitness);
			genotype_fprint(p->genotype, out);
			fprintf(out, "\n");
		}
	}
	else
	{
		fprintf(out, "%zu %f ", ga->generation, ga->best->phenotype->fitness);
		genotype_fprint(ga->best->genotype, out);
		fprintf(out, "\n");
	}
}

int
ga_fitness_fprint_header(struct ga *ga, FILE *out)
{
	return fprintf(out, "generation fitness_best fitness_avg "
		"fitness_stddev mutation_prob crossover_prob\n");
}

void
ga_fitness_fprint(struct ga *ga, FILE *out)
{
	fprintf(out, "%zu %f %f %f %zu %zu %f %zu %zu %f", ga->generation, ga->fitness_best,
		ga->fitness_avg, ga->fitness_stddev,
		ga->mutation_count, ga->mutation_possi,
		(double)ga->mutation_count / (double)ga->mutation_possi,
		ga->crossover_count, ga->crossover_possi,
		(double)ga->crossover_count / (double)ga->crossover_possi);
	dbg_assert(ga->best);
	if (ga->best)
	{
		fprintf(out, " ");
		genotype_fprint(ga->best->genotype, out);
	}
	fprintf(out, "\n");
}

void
ga_phenotype_fprint(struct ga *ga, FILE *out)
{
	dbg_assert(ga->best != NULL);
	if (ga->best)
	{
		phenotype_fprint(ga->best->phenotype, out);
	}
}

int
ga_parent_selection(struct ga *ga)
{
	int ret;

#if 0
	ret = ga->ops->parent_sel_proto(ga);
	if (ret)
	{
		dbg_error("%s\n", strerror(-ret));
		goto out;
	}
#endif

	ret = ga->ops->parent_sel_mech(ga);
	if (ret)
	{
		dbg_error("%s\n", strerror(-ret));
		goto out;
	}

out:
	return ret;
}

int
ga_reproduce(struct ga *ga)
{
	int ret;
	//dbg_assert(ga->children.end % 2 == 0);
	while (ga->parents.end)
	{
		struct individual child1, child2;
		struct individual parent1, parent2;
		double r;

		parent1 = DTABLE_DATA(&ga->parents, ga->parents.end - 1, struct individual);
		dtable_remove(&ga->parents, ga->parents.end - 1, struct individual);

		ret = individual_init_clone(&child1, parent1.genotype);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}

		if (ga->parents.end)
		{
			size_t j;

			j = prandi_range(ga->prand, 0, ga->parents.end);

			parent2 = DTABLE_DATA(&ga->parents, j, struct individual);
			dtable_remove(&ga->parents, j, struct individual);

			ret = individual_init_clone(&child2, parent2.genotype);
			if (ret)
			{
				dbg_error("%s\n", strerror(-ret));
				return ret;
			}

			++ga->crossover_possi;
			r = prandd(ga->prand);
			if (r <= ga->settings->crossover_prob)
			{
				++ga->crossover_count;
				genotype_crossover(child1.genotype,
					child2.genotype,
					ga->settings->crossover_points,
					ga->prand);
			}

			for (size_t k = 0; k < child2.genotype->features; ++k)
			{
				++ga->mutation_possi;
				r = prandd(ga->prand);
				if (r <= ga->settings->mutation_prob)
				{
					++ga->mutation_count;
					//dbg_debug("mutate!\n");
					genotype_mutate(child2.genotype, ga->prand, k);
				}
			}

			ret = dtable_insert(&ga->children, &child2,
				ga->children.end, struct individual);
			if (ret)
			{
				dbg_error("%s\n", strerror(-ret));
				return ret;
			}
		}
		//dbg_debug("%p %zu\n", parent1, i);
		//dbg_debug("%p %zu\n", parent2, i + 1);
		//dbg_debug("%p\n", parent1->genotype->ops);
		//dbg_debug("%p\n", parent2->genotype->ops);

		for (size_t k = 0; k < child1.genotype->features; ++k)
		{
			++ga->mutation_possi;
			r = prandd(ga->prand);
			if (r <= ga->settings->mutation_prob)
			{
				++ga->mutation_count;
				//dbg_debug("mutate!\n");
				genotype_mutate(child1.genotype, ga->prand, k);
			}
		}

		ret = dtable_insert(&ga->children, &child1,
			ga->children.end, struct individual);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}
	}

	return 0;
}

int
ga_run(struct ga *ga)
{
	int ret;

	if (ga->settings->fitness_file)
		ga_fitness_fprint_header(ga, ga->settings->fitness_file);
	ga->generation = 0;
	while (1)
	{
		//dbg_debug("generation %zu\n", ga->generation);
		ret = ga_children_develop(ga);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}

		ret = ga_children_test_fitness(ga);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}

		ret = ga_adult_selection(ga);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}

		if (ga->settings->population_file)
			ga_population_fprint(ga, ga->settings->population_file);
		if (ga->settings->fitness_file)
			ga_fitness_fprint(ga, ga->settings->fitness_file);
		/* FIXME: print stats */

		/* termination check */
		if (ga->generation >= ga->settings->generation_count)
			break;

		ret = ga_parent_selection(ga);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}

		ret = ga_reproduce(ga);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}
		++ga->generation;
	}

	return 0;
}

static int
individual_desc_cmp(const void *p1, const void *p2)
{
	const struct individual *a = (const struct individual *)p1;
	const struct individual *b = (const struct individual *)p2;

	if (a->phenotype->fitness < b->phenotype->fitness)
		return 1;
	else if (a->phenotype->fitness > b->phenotype->fitness)
		return -1;

	return 0;
}

static int
individual_asc_cmp(const void *p1, const void *p2)
{
	const struct individual *a = (const struct individual *)p1;
	const struct individual *b = (const struct individual *)p2;

	if (a->phenotype->fitness < b->phenotype->fitness)
		return -1;
	else if (a->phenotype->fitness > b->phenotype->fitness)
		return 1;

	return 0;
}

static void
adults_kill(struct ga *ga)
{
	if (ga->settings->elites_count)
		qsort(ga->adults.data, ga->adults.end,
			sizeof(struct individual),
			individual_desc_cmp);

	/* Kill adults - Check for zero since we do not have any elites
	 * in the first iteration. */
	while (ga->adults.end != 0 && ga->adults.end > ga->settings->elites_count)
	{
		individual_cleanup(&DTABLE_DATA(&ga->adults,
						ga->adults.end - 1,
						struct individual));
		dtable_remove(&ga->adults,
				ga->adults.end - 1,
				struct individual);
	}
}

static int
adults_copy_to_children(struct ga *ga)
{
	if (ga->settings->elites_count)
		qsort(ga->adults.data, ga->adults.end,
			sizeof(struct individual),
			individual_desc_cmp);

	while (ga->adults.end != 0 && ga->adults.end > ga->settings->elites_count)
	{
		int ret;

		ret = dtable_insert(&ga->children,
				&DTABLE_DATA(&ga->adults,
						ga->adults.end - 1,
						struct individual),
				ga->children.end,
				struct individual);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}
		dtable_remove(&ga->adults,
				ga->adults.end - 1,
				struct individual);
	}

	return 0;
}

int
ga_adult_sel_proto_full_gen_rep(struct ga *ga)
{
	adults_kill(ga);
	return 0;
}

int
ga_adult_sel_proto_over_pop(struct ga *ga)
{
	adults_kill(ga);
	return 0;
}

int
ga_adult_sel_proto_gen_mix(struct ga *ga)
{
	int ret;

	ret = adults_copy_to_children(ga);
	if (ret)
	{
		dbg_error("%s\n", strerror(-ret));
		return ret;
	}

	return 0;
}

int
ga_adult_sel_mech_best_fitness(struct ga *ga)
{
	qsort(ga->children.data, ga->children.end,
			sizeof(struct individual),
			individual_asc_cmp);

	while (ga->children.end != 0 &&
			ga->adults.end < ga->settings->adults_count)
	{
		int ret;

		ret = dtable_insert(&ga->adults,
				&DTABLE_DATA(&ga->children,
					ga->children.end - 1,
					struct individual),
				ga->adults.end,
				struct individual);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}

		dtable_remove(&ga->children,
				ga->children.end - 1,
				struct individual);

	}
	dbg_assert(ga->adults.end == ga->settings->adults_count);

	while (ga->children.end != 0)
	{
		individual_cleanup(&DTABLE_DATA(&ga->children,
						ga->children.end - 1,
						struct individual));
		dtable_remove(&ga->children,
				ga->children.end - 1,
				struct individual);
	}

	return 0;
}

int
ga_adult_sel_mech_stochastic(struct ga *ga)
{
	while (ga->adults.end < ga->settings->adults_count)
	{
		size_t idx;
		int ret;

		idx = prandi_range(ga->prand, 0, ga->children.end);;
		ret = dtable_insert(&ga->adults,
				&DTABLE_DATA(&ga->children,
						idx,
						struct individual),
				ga->adults.end,
				struct individual);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
			return ret;
		}
		dtable_remove(&ga->children,
				idx,
				struct individual);
	}
	dbg_assert(ga->adults.end == ga->settings->adults_count);

	while (ga->children.end != 0)
	{
		individual_cleanup(&DTABLE_DATA(&ga->children,
						ga->children.end - 1,
						struct individual));
		dtable_remove(&ga->children,
				ga->children.end - 1,
				struct individual);
	}

	return 0;
}
