#include <dtable.h>
#include <dbg.h>
#include <ga/ga.h>
#include <ga/individual.h>
#include <prand.h>

#include <math.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

/*
 * Spin the roulette wheel,
 * choose between fitness proportionate,
 * sigma scaling or boltzman scaling.
 */
static int
spin_roulette_wheel(struct ga *ga, double total)
{
	int ret;

	dbg_assert(ga->parents.end == 0);

	for (size_t j = 0; j < ga->settings->children_count; ++j)
	{
		double target, acc;

		target = prandd(ga->prand);
		acc = 0.0;
		for (size_t i = 0; i < ga->adults.end; ++i)
		{
			struct individual *p;

			p = &DTABLE_DATA(&ga->adults, i, struct individual);

			acc += (p->phenotype->scaled_fitness / total);
			if (acc > target || i == ga->adults.end - 1)
			{
				//dbg_debug("%zu parent %zu %f\n", ga->generation, i, p->phenotype->scaled_fitness / total);
				//dbg_debug("%zu %f\n", ga->generation, p->phenotype->fitness);
				ret = dtable_insert(&ga->parents,
						p,
						ga->parents.end,
						struct individual);
				if (ret)
				{
					dbg_error("%s\n", strerror(-ret));
				}
				break;
			}
		}
	}

	//dbg_debug("%zu %zu %zu\n", ga->parents.end, ga->settings->children_count, ga->settings->elites_count);
	dbg_assert(ga->parents.end == ga->settings->children_count);

	return 0;
}

int
fitness_proportionate_selection(struct ga *ga)
{
	double sum;

	sum = 0.0;
	for (size_t i = 0; i < ga->adults.end; ++i)
	{
		struct individual *p;

		p = &DTABLE_DATA(&ga->adults, i, struct individual);
		p->phenotype->scaled_fitness = p->phenotype->fitness;
		sum += p->phenotype->fitness;
	}

	return spin_roulette_wheel(ga, sum);
}

int
sigma_scale_selection(struct ga *ga)
{
	double sum;

	sum = 0.0;
	/* go through all adults and set scaled fitness */
	for (size_t i = 0; i < ga->adults.end; ++i)
	{
		struct individual *p;

		p = &DTABLE_DATA(&ga->adults, i, struct individual);
		/* double zero variance check */
		if (ga->fitness_stddev < 0.0001)
			p->phenotype->scaled_fitness = 1.0;
		else
			p->phenotype->scaled_fitness = 1.0 +
				(p->phenotype->fitness - ga->fitness_avg) /
				(2.0 * ga->fitness_stddev);

		/* give everyone a small chance to reproduce */
		if( p->phenotype->scaled_fitness < 0.0 )
			p->phenotype->scaled_fitness = 0.1;

		sum += p->phenotype->scaled_fitness;
	}

	return spin_roulette_wheel(ga, sum);
}

int
boltzman_scale(struct ga *ga, double temp)
{
	double avg_exp = 0.0;
	double sum;

	/* calculate the average fitness exponential in this generation */
	for (size_t i = 0; i < ga->adults.end; ++i)
	{
		struct individual *p;

		p = &DTABLE_DATA(&ga->adults, i, struct individual);
		avg_exp += exp(p->phenotype->fitness / temp);
	}

	avg_exp /= (double)ga->adults.end;
	sum = 0.0;
	/* go through them again and set the scaled fitness */
	for (size_t i = 0; i < ga->adults.end; ++i)
	{
		struct individual *p;

		p = &DTABLE_DATA(&ga->adults, i, struct individual);
		p->phenotype->scaled_fitness = exp(p->phenotype->fitness / temp) / avg_exp;
		sum += p->phenotype->scaled_fitness;
	}

	return spin_roulette_wheel(ga, sum);
}

/*
 * tournament selection
 * need settings.parent_selection_k and
 * settings.parent_selection_e to be set
 */
/* FIXME: do not modify parent array */
int
tournament_selection(struct ga *ga, size_t k, double e)
{
	int ret; /* error checking */

	//dbg_assert(n % 2 == 0);
	struct individual o[k]; /* the k individuals in the current tournament */

	for (size_t i = 0; i < ga->settings->children_count; ++i)
	{
		struct individual *winner;

		/* pick k random adults */
		for (size_t j = 0; j < k; ++j)
		{
			size_t r = prandi_range(ga->prand, 0, ga->adults.end);
			o[j] = DTABLE_DATA(&ga->adults,
						r,
						struct individual);
			/* take out selected adult,
			 * so it can't be chosen again
			 * in the same tournament.
			 */
			dtable_remove(&ga->adults,
					r,
					struct individual);

			/* pick winner, highest fitness wins
			 * with a probability of 1 - e,
			 * if highest doesn't win, random wins
			 */
		}
		double u = prandd(ga->prand);
		if (u < 1.0 - e)
		{
			size_t max = 0;

			for (size_t p = 1; p < k; ++p)
			{
				if(o[max].phenotype->fitness < o[p].phenotype->fitness)
				{
					max = p;
				}
			}

			winner = &o[max];
		}
		else
		{
			/* pick random winner */
			size_t r = prandi_range(ga->prand, 0, k);
			winner = &o[r];
		}

		/* copying best fitness in current tournament
		 * to temp mating pool
		 */
		ret = dtable_insert(&ga->parents,
				winner,
				ga->parents.end,
				struct individual);
		if (ret)
		{
			dbg_error("%s\n", strerror(-ret));
		}

		/* tournament selection over, put back the adults
		 * which was previously taken out.
		 */
		for (size_t j = 0; j < k; ++j)
		{
			ret = dtable_insert(&ga->adults,
					&o[j],
					ga->adults.end,
					struct individual);
			if (ret)
			{
				dbg_error("%s\n", strerror(-ret));
			}
		}
	}
	return 0;
}
