#include <util.h>
#include <dft.h>
#include <dbg.h>

#include <complex.h>
#include <fftw3.h>

#include <math.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

double
dft_window_hann(size_t n, size_t total)
{
	return 0.5 * (1 - cos((2 * M_PI * (double)n) / (double)(total - 1)));
}

double
dft_window_hamming(size_t n, size_t total)
{
	return 0.54 - 0.46 * cos((2 * M_PI * (double)n) / (double)(total - 1));
}

double
dft_window_rect(size_t n, size_t total)
{
	return 1.0;
}

int
dft_plan_init(struct dft_plan *plan, size_t window_size,
		double (*window_func)(size_t, size_t))
{
	int ret;

	if (window_func)
		plan->window_func = window_func;
	else
		plan->window_func = dft_window_rect;
	plan->window_size = window_size;
	plan->sample = fftw_malloc(plan->window_size * sizeof(*(plan->sample)));
	if (!plan->sample)
	{
		ret = -errno;
		dbg_error("%s\n", strerror(errno));
		goto sample_alloc_fail;
	}

	plan->transformed_size = (plan->window_size / 2) + 1;
	plan->transformed = fftw_malloc(plan->transformed_size * sizeof(*(plan->transformed)));
	if (!plan->transformed)
	{
		ret = -errno;
		dbg_error("%s\n", strerror(errno));
		goto transformed_alloc_fail;
	}

	plan->result = malloc(plan->transformed_size * sizeof(*(plan->result)));
	if (!plan->result)
	{
		ret = -errno;
		dbg_error("%s\n", strerror(errno));
		goto result_alloc_fail;
	}

	plan->plan = fftw_plan_dft_r2c_1d(plan->window_size, plan->sample,
		plan->transformed, FFTW_ESTIMATE);

	return 0;

result_alloc_fail:
	fftw_free(plan->transformed);
transformed_alloc_fail:
	fftw_free(plan->sample);
sample_alloc_fail:
	errno = -ret;
	return ret;
}

void
dft_init(struct dft_plan *dft)
{
	dft->sample_count = 0;
	dft->window_count = 0;
	memset(dft->result, 0, dft->transformed_size * sizeof(*(dft->result)));
}

void
dft_finish(struct dft_plan *dft)
{
	dbg_assert(!dft->sample_count);
	while (dft->sample_count)
	{
		dft_sample_add(dft, 0.0);
	}
}

void
dft_plan_cleanup(struct dft_plan *dft)
{
	fftw_destroy_plan(dft->plan);
	free(dft->result);
	fftw_free(dft->transformed);
	fftw_free(dft->sample);
}

static const struct
{
	const char * const name;
	double (*func)(size_t, size_t);
} dft_window_func_lookup[] =
{
	{ .name = "hann", .func = dft_window_hann },
	{ .name = "hamming", .func = dft_window_hamming },
	{ .name = "rect", .func = dft_window_rect },
};

double (*
dft_window_func_from_str(const char *func))(size_t, size_t)
{
	for (size_t i = 0; i < ARRAY_SIZE(dft_window_func_lookup); ++i)
	{
		if (!strcmp(func, dft_window_func_lookup[i].name))
			return dft_window_func_lookup[i].func;
	}

	return NULL;
}
