/**
 * \defgroup dtable Dynamic table.
 * \{
 * \file dtable.c
 * \brief Definitions for dynamic table.
 */
#include <dtable.h>

#include <limits.h>
#include <stdint.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

/**
 * \brief Clean up resources of \a dt.
 * \param dt Dynamic table to clean up after.
 * \return 0 on success.
 */
int
dtable_cleanup(struct dtable *dt)
{
	free(dt->data);
	dt->len = 0;
	dt->end = 0;
	dt->data = NULL;
	return 0;
}

/**
 * \internal
 * \brief Grows the size of the table in \a dt.
 * \param dt Pointer to dynamic table
 * \param size Size of an element.
 * \return 0 on success, -errno on failure
 */
static int
dtable_grow(struct dtable *dt, size_t size, size_t index)
{
	size_t new_len;
	void *new_data;

	new_len = dt->len;
	while (index >= new_len)
	{
		if (!new_len)
			new_len = DTABLE_DELTA;
		else if (new_len < SIZE_MAX / (size * 2))
			new_len = new_len * 2;
		else if (new_len < SIZE_MAX / size)
			new_len = SIZE_MAX / size;
		else
		{
			errno = EOVERFLOW;
			return -EOVERFLOW;
		}
	}

	new_data = realloc(dt->data, new_len * size);
	if (!new_data)
	{
		return -errno;
	}
	dt->len = new_len;
	dt->data = new_data;
	return 0;
}

/**
 * \brief Initialise \a dt to be a dynamic table of length \a len of elements
 * 	with size \a size.
 * \param dt Dynamic table to initialise.
 * \param len How many elements the table should initially have room for.
 * \param size Size of an element.
 * \return 0 on success, -errno on failure.
 */
int
_dtable_init(struct dtable *dt, size_t len, size_t size)
{
	int ret;

	ret = 0;

	if (len)
	{
		if (len > SIZE_MAX / size)
		{
			errno = EOVERFLOW;
			return -errno;
		}
		dt->data = malloc(len * size);
		if (!dt->data)
			ret = -errno;
		else
		{
			dt->len = len;
			dt->end = 0;
		}
	}
	else
	{
		dt->len = 0;
		dt->end = 0;
		dt->data = NULL;
	}

	return ret;
}

/**
 * \brief Insert \a data into \a dt as element \a index.
 * \param dt Dynamic table to insert into.
 * \param data Pointer to element to insert into table.
 * \param index Index of element to place \a data as.
 * \param size Size of an element.
 * \return 0 on success, -errno on failure.
 */
int
_dtable_insert(struct dtable *dt, const void *data, size_t index, size_t size)
{
	if (index == SIZE_MAX / size)
		return -EOVERFLOW;
	if (index >= dt->len)
	{
		int ret;

		ret = dtable_grow(dt, size, index);
		if (ret)
			return ret;
	}

	if (index < dt->end)
	{
		memmove(_dtable_ptr(dt, index + 1UL, size),
			_dtable_ptr(dt, index, size),
			size * (dt->end - index));
	}

	memcpy(_dtable_ptr(dt, index, size), data, size);

	/* FIXME: what when adding last element and ->end wrap around */
	if (index <= dt->end)
		++dt->end;
	else
		dt->end = index + 1;

	return 0;
}

/**
 * \brief Remove element at \a index from \a dt.
 * \param dt Dynamic table to remove from.
 * \param index Index of element to remove.
 * \param size Size of an element.
 */
void
_dtable_remove(struct dtable *dt, size_t index, size_t size)
{
	size_t index_max;

	index_max = dt->end - 1UL;
	if (index < index_max)
	{
		memmove(_dtable_ptr(dt, index, size),
			_dtable_ptr(dt, index + 1UL, size),
			size * (index_max - index));
	}
	--dt->end;
}

/**
 * \brief Copy an entire dynamic table from \a src to \a dst.
 * \param dst Dynamic table destination.
 * \param src Dynamic table source.
 * \param size Size of an element.
 * \return 0 on success, -errno on failure.
 */
int
_dtable_copy(struct dtable *dst, const struct dtable *src, size_t size)
{
	if (dst->len < src->end)
	{
		int ret;

		ret = dtable_grow(dst, size, src->end - 1);
		if (ret)
			return ret;
	}

	memcpy(dst->data, src->data, src->end * size);
	return 0;
}

/**
 * \}
 */
